/*
								+---------------------------------+
								|                                 |
								|       ***   Matice   ***        |
								|                                 |
								|    Copyright  L. Polok (Mr.)   |
								|                                 |
								|           matice.cpp            |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <math.h>
#include "Matice.h"

//////////////
//  Matrix  //
//////////////

void initmatrix(MATRIX * m)
{
	int i,j;
	for (j=0; j<4; j++) for (i=0; i<4; i++) (*m)[i][j] = (i==j);
}

void matmult(MATRIX * a, MATRIX * b, MATRIX * r)
{
	int i, j;

	MATRIX tmp;
	for (i=0; i<4; i++)
	{
		for (j=0; j<4; j++)
		tmp[i][j] = (*a)[0][j] * (*b)[i][0] + (*a)[1][j] * (*b)[i][1] +
					 (*a)[2][j] * (*b)[i][2] + (*a)[3][j] * (*b)[i][3];
	}
	memcpy (r,tmp,sizeof(MATRIX));
}

void rotatematrixx(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[2][1] = (float)-sin(angle);
	temp[2][2] = (float)cos(angle);
	temp[1][1] = (float)cos(angle);
	temp[1][2] = (float)sin(angle);
	matmult(m,&temp,m);
}

void rotatematrixy(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[0][0] = (float)cos(angle);
	temp[0][2] = (float)-sin(angle);
	temp[2][2] = (float)cos(angle);
	temp[2][0] = (float)sin(angle);
	matmult(m,&temp,m);
}

void rotatematrixz(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[0][0] = (float)cos(angle);
	temp[0][1] = (float)sin(angle);
	temp[1][1] = (float)cos(angle);
	temp[1][0] = (float)-sin(angle);
	matmult(m,&temp,m);
}

void translatematrix(MATRIX * m, float dx, float dy, float dz)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[3][0] = (float)dx;
	temp[3][1] = (float)dy;
	temp[3][2] = (float)dz;
	matmult(m,&temp,m);
}

void invertmatrix(MATRIX * src, MATRIX * inv)
{
	int i,j;
	float det, jdet;
	(*inv)[0][0] =   (*src)[1][1] * (*src)[2][2] - (*src)[1][2] * (*src)[2][1];
	(*inv)[0][1] = - (*src)[0][1] * (*src)[2][2] + (*src)[0][2] * (*src)[2][1];
	(*inv)[0][2] =   (*src)[0][1] * (*src)[1][2] - (*src)[0][2] * (*src)[1][1];
	det = (*inv)[0][0] * (*src)[0][0] + (*inv)[0][1] * (*src)[1][0] + (*inv)[0][2] * (*src)[2][0];
	if (det==0)
		return;
	(*inv)[1][0] = - (*src)[1][0] * (*src)[2][2] + (*src)[1][2] * (*src)[2][0];
	(*inv)[1][1] =   (*src)[0][0] * (*src)[2][2] - (*src)[0][2] * (*src)[2][0];
	(*inv)[1][2] = - (*src)[0][0] * (*src)[1][2] + (*src)[0][2] * (*src)[1][0];
	(*inv)[2][0] =   (*src)[1][0] * (*src)[2][1] - (*src)[1][1] * (*src)[2][0];
	(*inv)[2][1] = - (*src)[0][0] * (*src)[2][1] + (*src)[0][1] * (*src)[2][0];
	(*inv)[2][2] =   (*src)[0][0] * (*src)[1][1] - (*src)[0][1] * (*src)[1][0];
	jdet = 1 / det;
	for (j=0; j<3; j++) for (i=0; i<3; i++) (*inv)[j][i] *= jdet;
	(*inv)[3][0] = - (*src)[3][0]*(*inv)[0][0] - (*src)[3][1]*(*inv)[1][0] - (*src)[3][2]*(*inv)[2][0];
	(*inv)[3][1] = - (*src)[3][0]*(*inv)[0][1] - (*src)[3][1]*(*inv)[1][1] - (*src)[3][2]*(*inv)[2][1];
	(*inv)[3][2] = - (*src)[3][0]*(*inv)[0][2] - (*src)[3][1]*(*inv)[1][2] - (*src)[3][2]*(*inv)[2][2];
	(*inv)[0][3] = 0;
	(*inv)[1][3] = 0;
	(*inv)[2][3] = 0;
	(*inv)[3][3] = 1;
}