/*
                                +---------------------------------+
                                |                                 |
                                |     ***   Rasterizer   ***      |
                                |                                 |
                                |  Copyright   -tHE SWINe- 2002  |
                                |                                 |
                                |           Raster.cpp            |
                                |                                 |
                                +---------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "Load_3ds.h"
#include "Matrix.h"
#include "Pyramid.h"
#include "Font_Eng.h"

#include "Raster.h"

#define PI 3.14159265358979323846263279502884

struct Poly_Edge {                            
    int x, y, num, dxdy;            
};

struct PolyList {
    Face **face;
    float *face_z;
    int n_face_num;
};

static PolyList poly_tmp;
// painter's alg

static int n_Width, n_Height;
// vyska a sirka obrazovky

static float z_delta;
static Pyramid pyramid;
// klipovaci pyramida

static Poly_Edge left_edge_buff[7];
static Poly_Edge right_edge_buff[7];
static Poly_Edge *p_left_edge;
static Poly_Edge *p_right_edge;
static int n_left_edge_num, n_right_edge_num;
// edge

static float didx, didy, begi;
// interpolace

int Init_Engine(int n_New_Width, int n_New_Height)
{
    n_Width = n_New_Width;
    n_Height = n_New_Height;
    // nastav velikost obrazu

    z_delta = (float)fabs((float)n_Width / ((float)tan(90 * (PI / 180) / 2) * 2));
    // perspektivni zkresleni pro hel 90

    SetFOV((float)n_Width - 4, (float)n_Height - 4, z_delta, &pyramid);
    // klipovaci pyramida (je o trochu men kvli nepesnostem)

    return 1;
}

void Terminate_Engine()
{
    // zatm nic ...
}

int Add_Edge(Vertex *v_cur, Vertex *v_next, Poly_Edge *edge)
{
    float y1, y2;
    int num;

    y1 = (float)ceil(v_cur->y);
    y2 = (float)ceil(v_next->y);
    num = (int)(y2 - y1);

    if(!num)
        return 0;

    edge->dxdy = (int)((v_next->x - v_cur->x) / (v_next->y - v_cur->y) * 65536);
    edge->x = (int)(v_cur->x * 65536 + (y1 - v_cur->y) * edge->dxdy);
    edge->y = (int)y1;
    edge->num = num;
    // spot prvn x a y, sklon (dxdy) a poet obrazovch dek

    return 1;
}

int Create_Edges(Polygon *p_poly)
{
    int n_bottom, n_top, n_cur;
    float f_min_y;
    float f_max_y;

    n_left_edge_num = 0;
    n_right_edge_num = 0;
    // poty hran

    f_min_y =  0x10000;
    f_max_y = -0x10000;
    n_cur = 0;
    while(n_cur < p_poly->n_vertex_num) {
        if(p_poly->vertex[n_cur].y < f_min_y) {
            f_min_y = p_poly->vertex[n_cur].y;
            n_top = n_cur;
        }
        if(p_poly->vertex[n_cur].y > f_max_y) {
            f_max_y = p_poly->vertex[n_cur].y;
            n_bottom = n_cur;
        }
        n_cur ++;
    }
    // najde nejvrchnj a nejspodnj vrchol

    if((int)ceil(f_min_y) >= (int)ceil(f_max_y))
        return 0;
    // zkontroluje zda se vejde do rastru

    n_cur = n_top;
    while(n_cur != n_bottom) {
        if(Add_Edge(&p_poly->vertex[n_cur], &p_poly->vertex[(n_cur + 1) %
           p_poly->n_vertex_num], &right_edge_buff[n_right_edge_num]))
            n_right_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }
    // sjede pravou vtev

    n_cur = n_bottom;
    while(n_cur != n_top) {
        if(Add_Edge(&p_poly->vertex[(n_cur + 1) % p_poly->n_vertex_num],
           &p_poly->vertex[n_cur], &left_edge_buff[n_left_edge_num]))
            n_left_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }

    p_left_edge = left_edge_buff + n_left_edge_num - 1;
    p_right_edge = right_edge_buff;

    return 1;
}

int Compute_Interpolation_Consts(Polygon *pp)
{
    float x[2], y[2], illum[2];
    float a, b, c;
    
    x[0] = pp->vertex[1].x - pp->vertex[0].x;
    x[1] = pp->vertex[2].x - pp->vertex[0].x;

    y[0] = pp->vertex[1].y - pp->vertex[0].y;
    y[1] = pp->vertex[2].y - pp->vertex[0].y;
    
    illum[0] = pp->vertex[1].illum - pp->vertex[0].illum; 
    illum[1] = pp->vertex[2].illum - pp->vertex[0].illum;
    // vektory (od vrocholu 0 do 1, 2 pro kazdou souradnici)
    // vsimnete si, ze se pouziva 1/z a ne z

    a = y[0] * illum[1] - y[1] * illum[0];
    b = illum[0] * x[1] - illum[1] * x[0];
    c = x[0] * y[1] - x[1] * y[0];
    // cross product - normala polygonu

    if(c == 0) 
        return 0;
    
    didx = - a / c;
    didy = - b / c;
    // urceni potvor podle rovnice

    begi = pp->vertex[0].illum - pp->vertex[0].x * didx - pp->vertex[0].y * didy;
    // rychlejsi urceni - d / c

    return 1;
}

int DrawPolygon(Polygon *p, unsigned __int32 *video, unsigned __int32 color)
{
    int start_y, left_x, right_x;
    int dldx_fp, dldy_fp, k, r_k;
    float illum;
    int l, c;

    if(!Create_Edges(p)) 
        return 0;

    if(!Compute_Interpolation_Consts(p)) 
        return 0;

    start_y = p_right_edge->y;
    left_x = p_left_edge->x;
    right_x = p_right_edge->x;
    // zkopruje vlastnosti edge

    video += n_Width * start_y;
    // nastav video na prvn dku rastru

    illum = begi + didy * start_y;
    // z prvniho radku rastru

    dldx_fp = (int)(didx * 65536);
    dldy_fp = (int)(didy * 65536);
    // 16:16 fp

    while(1) {
        l = left_x >> 16;
        c = right_x >> 16;
        // pevod zptky na int

        k = (int)(illum * 65536) + c * dldx_fp;
        // 16/16 fixed point

        for(; c < l; c ++) {
            r_k = k >> 16;
            // zpet na cislo

            if(r_k > 255)
                r_k = 255;
            if(r_k < 0)
                r_k = 0;
            // kontroluje meze (to by tu byt nemelo, ale pak
            // to trochu na hranach "sumi")

            video[c] = ((((color & 0xff0000) * r_k) & 0xff000000) +
                (((color & 0x00ff00) * r_k) & 0x00ff0000) +
                (((color & 0x0000ff) * r_k) & 0x0000ff00)) >> 8;
            // mixne barvu
            
            k += dldx_fp;
            // dalsi pixel
        }
        // kresl vodorovny segment

        video += n_Width;
        illum += didy;
        // posun na dalsi ek obrazovky

        if(-- p_left_edge->num == 0) {
            if(p_left_edge -- == left_edge_buff)
                return 1;
            // pokud dojde na konec bufferu, je konec polygonu
            left_x = p_left_edge->x;
        } else 
            left_x += p_left_edge->dxdy;

        if(-- p_right_edge->num == 0) {
            p_right_edge ++;
            right_x = p_right_edge->x;
        } else 
            right_x += p_right_edge->dxdy;
    }

    return 1;
}

void ClearPolyList()
{
    poly_tmp.face_z = (float*)malloc(0);
    poly_tmp.face = (Face**)malloc(0);
    poly_tmp.n_face_num = 0;
    // vyist polylist
}

void FreePolyList()
{
    free((void*)poly_tmp.face);
    free((void*)poly_tmp.face_z);
    ClearPolyList();
    // uvoln polylist z pameti
}

void AddFaceToList(Face *a_face, float a_z)
{
    int i, j;

    for(i = 0; i < poly_tmp.n_face_num; i ++) {
        if(poly_tmp.face_z[i] < a_z) {
            poly_tmp.n_face_num ++;
            poly_tmp.face = (Face**)realloc(poly_tmp.face,
                poly_tmp.n_face_num * sizeof(Face*));
            poly_tmp.face_z = (float*)realloc(poly_tmp.face_z,
                poly_tmp.n_face_num * sizeof(float));
            // realloc
            for(j = poly_tmp.n_face_num - 1; j > i; j --) {
                poly_tmp.face[j] = poly_tmp.face[j - 1];
                poly_tmp.face_z[j] = poly_tmp.face_z[j - 1];
            }
            //
            poly_tmp.face[i] = a_face;
            poly_tmp.face_z[i] = a_z;
            //
            return;
        }
    }

    i = poly_tmp.n_face_num ++;
    poly_tmp.face = (Face**)realloc(poly_tmp.face,
        poly_tmp.n_face_num * sizeof(Face*));
    poly_tmp.face_z = (float*)realloc(poly_tmp.face_z,
        poly_tmp.n_face_num * sizeof(float));
    //
    poly_tmp.face[i] = a_face;
    poly_tmp.face_z[i] = a_z;
    // pid face do polylistu, ad podle vzdlenosti
}

void DrawObject(Matrix *m_object, Matrix *m_camera,
    Object *p_object, unsigned __int32 *video, unsigned __int32 color)
{
    float tx, ty, tz;
    float x, y, z;    
    Matrix m, im;
    Polygon poly;
    Vertex v[3];
    Face *cur_face, face;
    int i, j;

    face.vertex[0] = &v[0];
    face.vertex[1] = &v[1];
    face.vertex[2] = &v[2];
    // nastav vertexy pro nai doasnou face

    Matrix_Inverse(m_camera, &m);
    Matrix_Multiply(&m, m_object, &m);
    Matrix_Inverse(&m, &im);
    // spot transforman matice

    ClearPolyList();
    // vycisti polylist pro maliruv algorytmus

    for(i = 0; i < p_object->n_face_num; i ++){
        if((p_object->face[i].normal.c * im[3][2] +
           p_object->face[i].normal.a * im[3][0] +
           p_object->face[i].normal.b * im[3][1] +
           p_object->face[i].normal.d) > 0.1)
            continue;
        // byckface culling

        cur_face = &p_object->face[i];
        tz = 0;
        for(j = 0; j < 3; j ++){
            x = cur_face->vertex[j]->x;
            y = cur_face->vertex[j]->y;
            z = cur_face->vertex[j]->z;
            //
            tz += x * m[0][2] + y * m[1][2] + 
                 z * m[2][2] + m[3][2];
        }

        AddFaceToList(cur_face, tz);
    }
    // pidv facy do seznamu

    for(i = 0; i < poly_tmp.n_face_num; i ++){
        if((poly_tmp.face[i]->normal.c * im[3][2] +
           poly_tmp.face[i]->normal.a * im[3][0] +
           poly_tmp.face[i]->normal.b * im[3][1] +
           poly_tmp.face[i]->normal.d) > 0.1)
            continue;
        // backface culling

        cur_face = poly_tmp.face[i];
        for(j = 0; j < 3; j ++){
            x = cur_face->vertex[j]->x;
            y = cur_face->vertex[j]->y;
            z = cur_face->vertex[j]->z;
            //
            tx = x * m[0][0] + y * m[1][0] + 
                 z * m[2][0] +     m[3][0];
            ty = x * m[0][1] + y * m[1][1] + 
                 z * m[2][1] +     m[3][1];
            tz = x * m[0][2] + y * m[1][2] + 
                 z * m[2][2] +     m[3][2];
            //
            face.vertex[j]->x = tx;
            face.vertex[j]->y = ty;
            face.vertex[j]->z = tz;

            x = cur_face->vertex[j]->normal.x;
            y = cur_face->vertex[j]->normal.y;
            z = cur_face->vertex[j]->normal.z;
            tx = x * m[0][0] + y * m[1][0] + z * m[2][0];
            ty = x * m[0][1] + y * m[1][1] + z * m[2][1];
            tz = x * m[0][2] + y * m[1][2] + z * m[2][2];
            // natransformuje normlu

            x = (float)sqrt(tx * tx + ty * ty + tz * tz);
            // normalizuje ji

            face.vertex[j]->illum = (float)abs((int)(tz / x * 255));
        }

        poly = Clip_Face(&face, &pyramid);
        // proene face pyramidou

        if(poly.n_vertex_num) { // pokud je face vidt, bude to nenulov slo
            for(j = 0; j < poly.n_vertex_num; j ++) {
                if(poly.vertex[j].z < 0.1) {
                    poly.n_vertex_num = 0;
                    break;
                }
                // pokud je z - souadnice moc mal, zahodme polygon

                poly.vertex[j].x = (n_Width / 2) + (poly.vertex[j].x *
                    z_delta) / poly.vertex[j].z;
                poly.vertex[j].y = (n_Height / 2) + (poly.vertex[j].y *
                    z_delta) / poly.vertex[j].z;
                // perspektivn korekce
            }

            DrawPolygon(&poly, video, color);
            // vykresl polygon na obrazovku
        }
    }

    FreePolyList();
    // uvoln polylist z pamti
}
